from typing import Set

WHITELISTED_COLLECTIONS: Set = {"CompensationSummary", "EmploymentStatuses"}

WHITELISTED_FIELDS: Set = {
    "BaseRate",
    "BaseSalary",
    "PreviousBaseRate",
    "PreviousBaseSalary",
    "ChangeValue",
    "ChangePercent",
}

WHITELISTED_PAY_POLICY_CODES: Set = {"USA_CA_HNE", "USA_CA_HNE_4", "USA_CA_HNEWHSE", "USA_CA_HNEDRIVER"}



import os
from datetime import datetime, timedelta
from typing import ClassVar, Dict, List, Optional, Union

import attr
import backoff
import requests
import singer
from dayforce_client import Dayforce

from .utils import handle_rate_limit, is_fatal_code
from .whitelisting import WHITELISTED_COLLECTIONS, WHITELISTED_FIELDS, WHITELISTED_PAY_POLICY_CODES

LOGGER = singer.get_logger()


@attr.s
class DayforceStream(object):

    client: Dayforce = attr.ib(validator=attr.validators.instance_of(Dayforce))
    config: Dict = attr.ib(repr=False, validator=attr.validators.instance_of(Dict))
    config_path: Union[os.PathLike, str] = attr.ib()
    state: Dict = attr.ib(validator=attr.validators.instance_of(Dict))
    catalog: Optional[singer.catalog.Catalog] = attr.ib(
        validator=attr.validators.optional(attr.validators.instance_of(singer.catalog.Catalog)),
        repr=False,
        default=None,
    )
    catalog_path: Optional[Union[os.PathLike, str]] = attr.ib(default=None)

    @classmethod
    def from_args(cls, args, **kwargs):
        return cls(
            client=Dayforce(
                username=args.config.get("username"),
                password=args.config.get("password"),
                client_namespace=args.config.get("client_namespace"),
                dayforce_release=args.config.get("dayforce_release", 57),
                api_version=args.config.get("api_version", "V1"),
                test=args.config.get("test", False),
            ),
            config=args.config,
            config_path=args.config_path,
            catalog=getattr(args, "catalog", None),
            catalog_path=getattr(args, "catalog_path", None),
            state=args.state,
            **kwargs,
        )

    @staticmethod
    def get_schema(tap_stream_id: str, catalog: singer.catalog.Catalog) -> Dict:
        return catalog.get_stream(tap_stream_id).schema.to_dict()

    @staticmethod
    def get_bookmark(config: Dict, tap_stream_id: str, state: Dict, bookmark_properties: str) -> Optional[str]:
        bookmark = singer.bookmarks.get_bookmark(state, tap_stream_id, key=bookmark_properties)
        if bookmark is not None:
            return bookmark
        else:
            return config.get("start_date")


@attr.s
class DayforcePunchStream(DayforceStream):
    def sync(self):
        with singer.metrics.job_timer(job_type=f"sync_{self.tap_stream_id}"):
            with singer.metrics.record_counter(endpoint=self.tap_stream_id) as counter:
                start = singer.utils.strptime_to_utc(
                    self.get_bookmark(self.config, self.tap_stream_id, self.state, self.bookmark_properties)
                )
                new_bookmark = singer.utils.now()
                singer.bookmarks.write_bookmark(
                    state=self.state,
                    tap_stream_id=self.tap_stream_id,
                    key=self.bookmark_properties,
                    val=singer.utils.strftime(new_bookmark),
                )
                step = timedelta(days=6)
                while start < new_bookmark:
                    end = start + step - timedelta(seconds=1)
                    self._transform_records(start, end, counter)
                    start += step


@attr.s
class EmployeePunchesStream(DayforcePunchStream):
    tap_stream_id: ClassVar[str] = "employee_punches"
    key_properties: ClassVar[List[str]] = ["PunchXRefCode"]
    bookmark_properties: ClassVar[str] = "SyncTimestampUtc"
    replication_method: ClassVar[str] = "INCREMENTAL"

    @backoff.on_exception(
        backoff.expo, requests.exceptions.HTTPError, max_time=240, giveup=is_fatal_code, logger=LOGGER
    )
    @backoff.on_exception(
        backoff.expo, (requests.exceptions.ConnectionError, requests.exceptions.Timeout), max_time=240, logger=LOGGER
    )
    def _transform_records(self, start, end, counter):
        for _, record in self.client.get_employee_punches(
            filterTransactionStartTimeUTC=singer.utils.strftime(start),
            filterTransactionEndTimeUTC=singer.utils.strftime(end),
        ).yield_records():
            if record:
                record["SyncTimestampUtc"] = self.get_bookmark(
                    self.config, self.tap_stream_id, self.state, self.bookmark_properties
                )
                with singer.Transformer() as transformer:
                    transformed_record = transformer.transform(
                        data=record, schema=self.get_schema(self.tap_stream_id, self.catalog)
                    )
                    singer.write_record(
                        stream_name=self.tap_stream_id, time_extracted=singer.utils.now(), record=transformed_record
                    )
                    counter.increment()


@attr.s
class EmployeeRawPunchesStream(DayforcePunchStream):
    tap_stream_id: ClassVar[str] = "employee_raw_punches"
    key_properties: ClassVar[List[str]] = ["RawPunchXRefCode"]
    bookmark_properties: ClassVar[str] = "SyncTimestampUtc"
    replication_method: ClassVar[str] = "INCREMENTAL"

    @backoff.on_exception(
        backoff.expo, requests.exceptions.HTTPError, max_time=240, giveup=is_fatal_code, logger=LOGGER
    )
    @backoff.on_exception(
        backoff.expo, (requests.exceptions.ConnectionError, requests.exceptions.Timeout), max_time=240, logger=LOGGER
    )
    def _transform_records(self, start, end, counter):
        for _, record in self.client.get_employee_raw_punches(
            filterTransactionStartTimeUTC=singer.utils.strftime(start),
            filterTransactionEndTimeUTC=singer.utils.strftime(end),
        ).yield_records():
            if record:
                record["SyncTimestampUtc"] = self.get_bookmark(
                    self.config, self.tap_stream_id, self.state, self.bookmark_properties
                )
                with singer.Transformer() as transformer:
                    transformed_record = transformer.transform(
                        data=record, schema=self.get_schema(self.tap_stream_id, self.catalog)
                    )
                    singer.write_record(
                        stream_name=self.tap_stream_id, time_extracted=singer.utils.now(), record=transformed_record
                    )
                    counter.increment()


@attr.s
class EmployeesStream(DayforceStream):
    tap_stream_id: ClassVar[str] = "employees"
    key_properties: ClassVar[List[str]] = ["XRefCode"]
    bookmark_properties: ClassVar[str] = "SyncTimestampUtc"
    replication_method: ClassVar[str] = "INCREMENTAL"

    def whitelist_sensitive_info(self, data: Dict) -> Dict:
        for collection in WHITELISTED_COLLECTIONS:
            if data.get(collection, {}).get("Items") is not None:
                for i, item in enumerate(data.get(collection, {}).get("Items")):
                    if (
                        item.get("PayPolicy") is None
                        or item.get("PayPolicy").get("XRefCode") not in WHITELISTED_PAY_POLICY_CODES
                    ):
                        for field in WHITELISTED_FIELDS:
                            data[collection]["Items"][i].pop(field, None)

        return data

    @backoff.on_exception(
        backoff.expo, requests.exceptions.HTTPError, max_time=240, giveup=is_fatal_code, logger=LOGGER
    )
    @backoff.on_exception(
        backoff.expo, (requests.exceptions.ConnectionError, requests.exceptions.Timeout), max_time=240, logger=LOGGER
    )
    def _transform_records(self, start, end, counter):
        for _, record in self.client.get_employees(
            filterUpdatedStartDate=singer.utils.strftime(start), filterUpdatedEndDate=singer.utils.strftime(end)
        ).yield_records():
            if record:

                details = handle_rate_limit(
                    func=self.client.get_employee_details(
                        xrefcode=record.get("XRefCode"),
                        expand="WorkAssignments,Contacts,EmploymentStatuses,Roles,EmployeeManagers,CompensationSummary,Locations,LastActiveManagers",
                    ),
                    logger=LOGGER,
                ).get("Data")

                schedules = handle_rate_limit(
                    func=self.client.get_employee_schedules(
                        xrefcode=record.get("XRefCode"),
                        filterScheduleStartDate=singer.utils.strftime(start),
                        filterScheduleEndDate=singer.utils.strftime(end),
                        expand="Activities,Breaks,Skills,LaborMetrics",
                    ),
                    logger=LOGGER,
                ).get("Data")

                if details.get("XRefCode") is not None:
                    details["SyncTimestampUtc"] = self.get_bookmark(
                        self.config, self.tap_stream_id, self.state, self.bookmark_properties
                    )
                    details = self.whitelist_sensitive_info(data=details)
                    if schedules is not None:
                        details["Schedules"] = schedules
                    with singer.Transformer() as transformer:
                        transformed_record = transformer.transform(
                            data=details, schema=self.get_schema(self.tap_stream_id, self.catalog)
                        )
                        singer.write_record(
                            stream_name=self.tap_stream_id, time_extracted=singer.utils.now(), record=transformed_record
                        )
                        counter.increment()

    def sync(self):
        with singer.metrics.job_timer(job_type=f"sync_{self.tap_stream_id}"):
            with singer.metrics.record_counter(endpoint=self.tap_stream_id) as counter:
                start = singer.utils.strptime_to_utc(
                    self.get_bookmark(self.config, self.tap_stream_id, self.state, self.bookmark_properties)
                )
                new_bookmark = singer.utils.now()
                singer.bookmarks.write_bookmark(
                    state=self.state,
                    tap_stream_id=self.tap_stream_id,
                    key=self.bookmark_properties,
                    val=singer.utils.strftime(new_bookmark),
                )
                self._transform_records(start, new_bookmark, counter)


@attr.s
class PaySummaryReportStream(DayforceStream):
    tap_stream_id: ClassVar[str] = "pay_summary_report"
    key_properties: ClassVar[List[str]] = []
    bookmark_properties: ClassVar[List[str]] = []
    replication_method: ClassVar[str] = "FULL_TABLE"
    date_param_fmt: ClassVar[str] = "%m/%d/%Y %I:%M:%S %p"

    @backoff.on_exception(
        backoff.expo, requests.exceptions.HTTPError, max_time=240, giveup=is_fatal_code, logger=LOGGER
    )
    @backoff.on_exception(
        backoff.expo, (requests.exceptions.ConnectionError, requests.exceptions.Timeout), max_time=240, logger=LOGGER
    )
    def _transform_records(
        self, start: datetime, end: datetime, counter: singer.metrics.Counter, time_extracted: datetime
    ):
        report_params = {
            "003cd1ea-5f11-4fe8-ae9c-d7af1e3a95d6": singer.utils.strftime(start, format_str=self.date_param_fmt),
            "b03cd1ea-5f11-4fe8-ae9c-d7af1e3a95d6": singer.utils.strftime(end, format_str=self.date_param_fmt),
        }
        rows_returned = 0
        for _, row in self.client.get_report(xrefcode="pay_summary_report", **report_params).yield_report_rows(
            limit=(500, 3600)
        ):
            if row:
                rows_returned += 1
                with singer.Transformer() as transformer:
                    transformed_record = transformer.transform(
                        data=row, schema=self.get_schema(self.tap_stream_id, self.catalog)
                    )
                    singer.write_record(
                        stream_name=self.tap_stream_id, record=transformed_record, time_extracted=time_extracted
                    )
                    counter.increment()

        if 18000 <= rows_returned < 20000:
            LOGGER.warning("Approaching maximum row limit of 20,000. Consider making request window smaller.")
        elif rows_returned >= 20000:
            LOGGER.error("Hit maximum row limit of 20,000. Make request window smaller for Pay Summary Report.")

    def sync(self):
        with singer.metrics.job_timer(job_type=f"sync_{self.tap_stream_id}"):
            with singer.metrics.record_counter(endpoint=self.tap_stream_id) as counter:
                start = singer.utils.strptime_to_utc(self.config.get("start_date"))
                new_bookmark = singer.utils.now()
                step = timedelta(days=6)
                while start < new_bookmark:
                    end = start + step - timedelta(seconds=1)
                    LOGGER.info(f"Running Pay Summary Report for {start} to {end} ..")
                    self._transform_records(start=start, end=end, counter=counter, time_extracted=new_bookmark)
                    start += step



import json
import os
import sys

import rollbar
import singer

from .streams import EmployeePunchesStream, EmployeeRawPunchesStream, EmployeesStream, PaySummaryReportStream
from .utils import load_schema, parse_args

AVAILABLE_STREAMS = {EmployeePunchesStream, EmployeeRawPunchesStream, EmployeesStream, PaySummaryReportStream}

LOGGER = singer.get_logger()

try:
    ROLLBAR_ACCESS_TOKEN = os.environ["ROLLBAR_ACCESS_TOKEN"]
    ROLLBAR_ENVIRONMENT = os.environ["ROLLBAR_ENVIRONMENT"]
except KeyError:
    LOGGER.info("Required env vars for Rollbar logging not found. Rollbar logging disabled..")
    log_to_rollbar = False
else:
    rollbar.init(ROLLBAR_ACCESS_TOKEN, ROLLBAR_ENVIRONMENT)
    log_to_rollbar = True


def discover(args, select_all=False):
    LOGGER.info("Starting discovery..")

    catalog = {"streams": []}
    for stream in AVAILABLE_STREAMS:
        schema = load_schema(stream.tap_stream_id)
        catalog_entry = {
            "stream": stream.tap_stream_id,
            "tap_stream_id": stream.tap_stream_id,
            "schema": schema,
            "metadata": singer.metadata.get_standard_metadata(
                schema=schema,
                key_properties=stream.key_properties,
                valid_replication_keys=stream.bookmark_properties,
                replication_method=stream.replication_method,
            ),
        }

        if select_all is True:
            catalog_entry["metadata"][0]["metadata"]["selected"] = True
        catalog["streams"].append(catalog_entry)

    print(json.dumps(catalog, indent=2))
    LOGGER.info("Finished discovery..")


def sync(args):
    LOGGER.info("Starting sync..")

    selected_streams = {catalog_entry.stream for catalog_entry in args.catalog.get_selected_streams(args.state)}
    LOGGER.info(f"Selected Streams: {selected_streams}")

    for available_stream in AVAILABLE_STREAMS:
        stream = available_stream.from_args(args)
        if stream.tap_stream_id in selected_streams:
            LOGGER.info(f"Starting sync for Stream {stream.tap_stream_id}..")
            singer.bookmarks.set_currently_syncing(state=stream.state, tap_stream_id=stream.tap_stream_id)
            singer.write_state(stream.state)
            singer.write_schema(
                stream_name=stream.tap_stream_id,
                schema=stream.get_schema(tap_stream_id=stream.tap_stream_id, catalog=stream.catalog),
                key_properties=stream.key_properties,
            )
            stream.sync()
            singer.bookmarks.set_currently_syncing(state=stream.state, tap_stream_id=None)
            singer.write_state(stream.state)


def _main():
    args = parse_args(required_config_keys={"username", "password", "client_namespace", "start_date"})
    if args.discover:
        discover(args, select_all=args.select_all)
    elif not args.catalog:
        raise RuntimeError("Catalog file must be supplied during Sync.")
    else:
        sync(args)


def main():
    try:
        _main()
    except Exception:
        if log_to_rollbar is True:
            LOGGER.info("Reporting exception info to Rollbar..")
            rollbar.report_exc_info()
        LOGGER.exception(msg="Uncaught Exception..")
        sys.exit(1)


if __name__ == "__main__":
    main()



import argparse
import json
import logging
import os
import time
from typing import Callable, Dict, Set

import requests
import singer
from dayforce_client.response import DayforceResponse


def get_abs_path(path: str) -> str:
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), path)


def load_schema(tap_stream_id: str, schema_path: str = "schemas") -> Dict:
    path = get_abs_path(schema_path)
    return singer.utils.load_json(f"{path}/{tap_stream_id}.json")


def is_fatal_code(e: requests.exceptions.RequestException) -> bool:
    """Helper function to determine if a Requests reponse status code
    is a "fatal" status code. If it is, the backoff decorator will giveup
    instead of attemtping to backoff."""
    return 400 <= e.response.status_code < 500 and e.response.status_code != 429


def load_json(path: str) -> Dict:
    with open(path) as fil:
        return json.load(fil)


def parse_args(required_config_keys: Set[str]) -> argparse.Namespace:
    """Parse standard command-line args.
    Parses the command-line arguments mentioned in the SPEC and the
    BEST_PRACTICES documents:
    -c,--config     Config file
    -s,--state      State file
    -d,--discover   Run in discover mode
    -p,--properties Properties file: DEPRECATED, please use --catalog instead
    -a,--select-all Selects all streams in the Catalog for replication.
    --catalog       Catalog file
    Returns the parsed args object from argparse. For each argument that
    point to JSON files (config, state, properties), we will automatically
    load and parse the JSON file.
    """
    parser = argparse.ArgumentParser()

    parser.add_argument("-c", "--config", help="Config file", required=True)

    parser.add_argument("-s", "--state", help="State file")

    parser.add_argument("--catalog", help="Catalog file")

    parser.add_argument("-d", "--discover", action="store_true", help="Do schema discovery")

    parser.add_argument(
        "-a", "--select-all", action="store_true", help="Selects all streams in the Catalog for replication."
    )

    args = parser.parse_args()
    if args.config:
        setattr(args, "config_path", args.config)
        args.config = load_json(args.config)
    if args.state:
        setattr(args, "state_path", args.state)
        args.state = load_json(args.state)
    else:
        args.state = {}
    if args.catalog:
        setattr(args, "catalog_path", args.catalog)
        args.catalog = singer.Catalog.load(args.catalog)

    check_config(args.config, required_config_keys)

    return args


def check_config(config: Dict, required_config_keys: Set[str]):
    missing_keys = [key for key in required_config_keys if key not in config]
    if missing_keys:
        raise Exception("Config is missing required keys: {}".format(missing_keys))


def handle_rate_limit(func: Callable, logger: logging.Logger) -> DayforceResponse:
    try:
        response = func
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 429:
            sleep_time = int(e.response.headers["Retry-After"]) + 1
            logger.info(f"Rate limit reached. Retrying in {sleep_time} seconds..")
            time.sleep(sleep_time)
            response = func
        else:
            raise
    finally:
        return response



from setuptools import find_packages, setup


def get_version():
    version = {}
    with open("tap_dayforce/version.py") as fp:
        exec(fp.read(), version)
    return version["__version__"]


with open("README.md", "r") as f:
    readme = f.read()


setup(
    name="tap_dayforce",
    author="David Wallace",
    author_email="david.wallace@goodeggs.com",
    version=get_version(),
    url="https://github.com/goodeggs/tap-dayforce",
    description="Singer.io tap for extracting data from Dayforce REST API v1",
    long_description=readme,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Topic :: Software Development",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    keywords="singer tap python dayforce",
    license="GPLv3",
    packages=find_packages(exclude=["tests"]),
    package_data={"tap_dayforce": ["schemas/*.json"]},
    install_requires=["dayforce-client==0.2.2", "singer-python==5.9.0", "backoff==1.8.0", "rollbar==0.14.7"],
    python_requires=">=3.6",
    entry_points={"console_scripts": ["tap-dayforce = tap_dayforce:main"]},
)
